/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.datamodel;

import java.util.ArrayList;

import javax.jcr.PathNotFoundException;
import javax.jcr.RepositoryException;

/**
 * Created by The eXo Platform SARL        .
 *
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov</a>
 * @version $Id: InternalQPath.java 5622 2006-05-19 16:09:57Z peterit $
 */

public class InternalQPath {

    private Entry[] names;
	public InternalQPath() {
		names = new Entry[0];
	}
	
	public void addEntry(String namespace, String name, int index) {
    addEntry(new Entry(namespace, name, index));
	}
	
	private void addEntry(Entry entry) {
		if(entry.getIndex() < 1)
			entry.setIndex(1);
		Entry[] newNames = new Entry[names.length+1];
        for(int i=0; i<names.length; i++)
        	newNames[i] = names[i];
        newNames[names.length] = entry;
        names = newNames;
	}	
    
    /**
     * @return parent path
     * @throws PathNotFoundException if path could not have parent - i.e. root path
     */
    public InternalQPath makeParentPath() 
    throws PathNotFoundException, IllegalPathException {
    	return makeAncestorPath(1);
    }
    
    /**
     * Makes ancestor path by relative degree (For ex relativeDegree == 1 
     * means parent path etc)
     * @param relativeDegree
     * @return
     * @throws PathNotFoundException
     */
    public InternalQPath makeAncestorPath(int relativeDegree) 
    throws PathNotFoundException, IllegalPathException {
    	InternalQPath path = new InternalQPath();
        if(relativeDegree > getLength() || getLength() <= 1)
           throw new IllegalPathException("Relative degree "+relativeDegree+" is more than depth for "+getAsString());
        else {
          for(int i=0; i<(getLength() - relativeDegree); i++) 
          	path.addEntry(names[i].getNamespace(), names[i].getName(), names[i].getIndex());
        }
        return path;
    }
    
    public Entry[] getRelPath(int relativeDegree) throws IllegalPathException {
      
      if(relativeDegree > getLength() || getLength() <= 1)
        throw new IllegalPathException("Relative degree "+relativeDegree+" is more than depth for "+getAsString());
      
      ArrayList entries = new ArrayList();
      for (int i = 0; i < names.length - relativeDegree; i++)
        entries.add(names[i]);
      return (Entry[]) entries.toArray();
    }

    
    /**
     * @return array of its path's names
     */
    public Entry[] getEntries() {
    	return names;
    }

    /**
     * @return depth of this path calculates as size of names array - 1.
     * For ex root's depth=0 etc.
     */
    public int getDepth() {
        return getLength()-1;
    }
    
    /**
     * @param anotherPath
     * @param childOnly if == true only direct children of the path will be 
     * taking in account 
     * @return if this path is descendant of another one 
     */
    public boolean isDescendantOf(InternalQPath anotherPath, boolean childOnly) {
        int depthDiff = getDepth() - anotherPath.getDepth();
        if(depthDiff <=0 || (childOnly && depthDiff!=1))
            return false;

        InternalQName[] anotherNames = anotherPath.getEntries();
        for(int i=0; i<anotherNames.length; i++) {
            boolean result = anotherNames[i].equals(names[i]);
            if(!result)
                return false;
        }
        return true;
    }
    

    /**
     * @return last name of this path
     */
    public InternalQName getName() {
    	InternalQName name = (InternalQName)names[getLength()-1];
    	return new InternalQName(name.getNamespace(), name.getName()); 
    }
    
    /**
     * @return index 
     */
    public int getIndex() {
    	return names[getLength()-1].getIndex();
    }

    /**
     * @return length of names array
     */
    public int getLength() {
        return names.length;
    }

    public String getAsString() {
        String str = "";
        for(int i=0; i<getLength(); i++) {
            str+=names[i].getAsString();
        }
        return str;
    }
    
    public boolean equals(Object o) {
    	if(!(o instanceof InternalQPath))
    		return false;
    	else
    		return getAsString().equals(((InternalQPath)o).getAsString());
    }
    
    // Factory methods ---------------------------

	/**
	 * Parses string and make internal path from it
	 * @param qPath
	 * @return
	 * @throws RepositoryException - if string is invalid
	 */
	public static InternalQPath parse(String qPath) throws IllegalPathException {
		InternalQPath path = new InternalQPath();
	  	if(qPath == null)
	    	throw new IllegalPathException("Bad internal path '"+qPath+"'");
	
	    if(qPath.length() < 2 || !qPath.startsWith("[]"))
	        throw new IllegalPathException("Bad internal path '"+qPath+"'");
	
	    int uriStart = 0;
	    while(uriStart>=0) {
	    	
	        uriStart = qPath.indexOf("[", uriStart);
	        
	        int uriFinish = qPath.indexOf("]", uriStart);
	        String uri = qPath.substring(uriStart+1, uriFinish);
	        
	        int tmp = qPath.indexOf("[", uriFinish); // next token
	        if(tmp == -1) {
	            tmp = qPath.length();
	            uriStart = -1;
	        } else
	            uriStart = tmp;
	        
	        String localName = qPath.substring(uriFinish+1, tmp);
	        int index = 0;
	        int ind = localName.indexOf(":");
	        if(ind != -1) { // has index
	            index = Integer.parseInt(localName.substring(ind+1));
	            localName = localName.substring(0, ind);
	        } else {
	        	if(uriStart > -1)
	        		throw new IllegalPathException("Bad internal path '"+qPath+"' each intermediate name should have index");
	        }
	
	        path.addEntry(uri, localName, index);
	    }
	    return path;
	}

	/**
	 * Makes child path from existed path and child name. Assumed that parent path belongs
	 * to node so it should have some index. If not sets index=1 automatically 
	 * @param parent path
	 * @param name child name
	 * @return new InternalQPath
	 */
	public static InternalQPath makeChildPath(InternalQPath parent, String name) 
	throws IllegalPathException{
		InternalQPath path = new InternalQPath();
	    for(int i=0; i<parent.getLength(); i++)
	    	path.addEntry((Entry)parent.getEntries()[i]);
	    path.addEntry(path.parseEntry(name));
	    return path;
	}
	
	
    private Entry parseEntry(String entry) throws IllegalPathException {
    	
    	if(!entry.startsWith("["))
    		throw new IllegalPathException("Invalid QPath Entry '"+entry+"' Should start of '['");
    	int uriStart = 0;
        int uriFinish = entry.indexOf("]", uriStart);
    	if(uriFinish == -1)
    		throw new IllegalPathException("Invalid QPath Entry '"+entry+"' No closed ']'");
        String uri = entry.substring(uriStart+1, uriFinish);

        String localName = entry.substring(uriFinish+1, entry.length());

        int index = 1; // default value
        int ind = localName.indexOf(":");
        if(ind > 1) {
            index = Integer.parseInt(localName.substring(ind+1));
            localName = localName.substring(0, ind);
        }

        return new Entry(uri, localName, index);
    }
	


	public static class Entry extends InternalQName {
		
		private int index; 
		
	    public Entry(String namespace, String name, int index) {
			super(namespace, name);
			this.index = index;
		}
	    
		public int getIndex() {
			return index;
		}
		public void setIndex(int index) {
			this.index = index;
		}
		
		public boolean equals(Object obj) {
			if(!(obj instanceof Entry))
				return false;
			Entry another = (Entry) obj;
			return (namespace.equals(another.getNamespace()) 
					&& name.equals(another.getName()) && index == another.getIndex());
		}
		
	    public String getAsString() {
	        return "["+namespace+"]"+name+":"+index;
	    }

    }
    
}
