/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl;

import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import javax.jcr.NamespaceRegistry;
import javax.jcr.RepositoryException;
import javax.jcr.nodetype.NodeTypeManager;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.config.RepositoryConfigurationException;
import org.exoplatform.services.jcr.config.RepositoryEntry;
import org.exoplatform.services.jcr.config.WorkspaceEntry;
import org.exoplatform.services.jcr.impl.core.LocationFactory;
import org.exoplatform.services.jcr.impl.core.NamespaceRegistryImpl;
import org.exoplatform.services.jcr.impl.core.RepositoryImpl;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.core.UUIDGenerator;
import org.exoplatform.services.jcr.impl.core.WorkspaceInitializer;
import org.exoplatform.services.jcr.impl.core.access.AccessManager;
import org.exoplatform.services.jcr.impl.core.lock.LockManager;
import org.exoplatform.services.jcr.impl.core.nodetype.NodeTypeManagerImpl;
import org.exoplatform.services.jcr.impl.core.observation.ObservationManagerRegistry;
import org.exoplatform.services.jcr.impl.core.query.QueryHandler;
import org.exoplatform.services.jcr.impl.core.query.QueryHandlerContext;
import org.exoplatform.services.jcr.impl.core.query.QueryManagerFactory;
import org.exoplatform.services.jcr.impl.core.query.SearchManager;
import org.exoplatform.services.jcr.impl.core.value.ValueFactoryImpl;
import org.exoplatform.services.jcr.impl.dataflow.persistent.BatcheableWorkspaceDataManager;
import org.exoplatform.services.jcr.impl.dataflow.persistent.WorkspaceDataManager;
import org.exoplatform.services.jcr.impl.dataflow.persistent.WorkspaceStorageCacheImpl;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL 11.05.2006
 * 
 * @author <a href="mailto:peter.nedonosko@exoplatform.com.ua">Peter Nedonosko</a>
 * @version $Id: RepositoryInitializer.java 5834 2006-05-29 12:05:59Z peterit $
 */
public class RepositoryInitializer {

  static protected Log log = ExoLogger.getLogger("jcr.WorkspaceStorageInitializer");
  
  private RepositoryContainer container;

  private RepositoryEntry     config;
  
  private BatcheableWorkspaceDataManager systemDataManager = null;

  private static class WorkspaceOrderComparator implements Comparator {
    private String sysWs;
    
    private WorkspaceOrderComparator(String sysWs) {
      this.sysWs = sysWs;
    }
    
    public int compare(Object o1, Object o2) {
      String n1 = ((WorkspaceEntry) o1).getName();
      return n1.equals(sysWs)?-1:0;
    }
  }  
  
  RepositoryInitializer(RepositoryContainer container, RepositoryEntry config)
      throws RepositoryException, RepositoryConfigurationException {
    this.container = container;
    this.config = config;
    
    container.registerComponentImplementation(UUIDGenerator.class);
    
    registerWorkspaces();
    registerRepositoryEnviroment();
  }

  private void registerRepositoryEnviroment() throws RepositoryConfigurationException {
    
    container.registerComponentImplementation(NamespaceRegistryImpl.class);
    container.registerComponentImplementation(LocationFactory.class);
    container.registerComponentImplementation(ValueFactoryImpl.class);
    container.registerComponentImplementation(NodeTypeManagerImpl.class);
    container.registerComponentImplementation(AccessManager.class);
    try {
      container.registerComponentImplementation(Class.forName(config.getAuthenticationPolicy()));
    } catch (ClassNotFoundException e) {
      throw new RepositoryConfigurationException(
          "Class not found for repository authentication policy: " + e);
    }
  }
  
  private void registerWorkspaceEnviroment(WorkspaceContainer wsContainer, WorkspaceEntry wsConfig) throws RepositoryConfigurationException {
    
    wsConfig.setUniqueName(container.getName() + "_" + wsConfig.getName());

    wsContainer.registerComponentInstance(wsConfig);
    try {
      wsContainer.registerComponentImplementation(Class.forName(wsConfig.getContainer().getType()));
    } catch (ClassNotFoundException e) {
      throw new RepositoryConfigurationException(
          "Class not found for workspace data container " + wsConfig.getUniqueName() + ": " + e);
    }

    wsContainer.registerComponentImplementation(WorkspaceStorageCacheImpl.class);
    wsContainer.registerComponentImplementation(BatcheableWorkspaceDataManager.class);   
    wsContainer.registerComponentImplementation(ObservationManagerRegistry.class);
    wsContainer.registerComponentImplementation(LockManager.class);
    if (wsConfig.getQueryHandler() != null) {
      try {
        Class qh = Class.forName(wsConfig.getQueryHandler().getType());
        wsContainer.registerComponentImplementation(qh);
      } catch (ClassNotFoundException e) {
        throw new RepositoryConfigurationException(
            "Class not found for workspace query handler, container " + wsConfig.getUniqueName() + ": " + e);
      }
      wsContainer.registerComponentImplementation(SearchManager.class);
      wsContainer.registerComponentImplementation(QueryManagerFactory.class);
    }
    wsContainer.registerComponentImplementation(WorkspaceInitializer.class);
  }

  private void registerWorkspaces() throws RepositoryException, RepositoryConfigurationException {
    List wsEntries = config.getWorkspaceEntries();
    Collections.sort(wsEntries, new WorkspaceOrderComparator(config.getSystemWorkspaceName()));
    for (int i = 0; i < wsEntries.size(); i++) {
      registerWorkspace((WorkspaceEntry) wsEntries.get(i), i == 0); // i == 0 - it's system workspace      
    }
  }  
  
  private void registerWorkspace(WorkspaceEntry wsConfig, boolean isSystem) 
  throws RepositoryException, RepositoryConfigurationException {
    WorkspaceContainer workspaceContainer = new WorkspaceContainer(container, wsConfig);
    
    registerWorkspaceEnviroment(workspaceContainer, wsConfig);
    
    container.registerComponentInstance(wsConfig.getName(), workspaceContainer);
    
    BatcheableWorkspaceDataManager wsDataManager = 
      (BatcheableWorkspaceDataManager) workspaceContainer.getComponentInstanceOfType(
          BatcheableWorkspaceDataManager.class);
    if (isSystem) {
      // system workspace
      systemDataManager = wsDataManager;
      container.registerComponentInstance(systemDataManager);
      // TODO system session
      // [PN] 16.05.06
      //SessionImpl sysSession = new SessionImpl(
      //    wsConfig.getName(), AccessManager.systemCredentials(), workspaceContainer);
      //container.registerComponentInstance(sysSession);
    }
    wsDataManager.setSystemDataManager(systemDataManager);
  }
  
  /**
   * Register particular workspace using WorkspaceEntry configuration.
   * USECASE: For a workspace creation in runtime.
   * @param {@link WorkspaceEntry wsConfig}
   */
  public void registerWorkspace(WorkspaceEntry wsConfig) throws RepositoryException, RepositoryConfigurationException {
    registerWorkspace(wsConfig, false);
  }
  
  void init(RepositoryImpl repository) throws RepositoryException, RepositoryConfigurationException {
    List wsEntries = config.getWorkspaceEntries();
    Collections.sort(wsEntries, new WorkspaceOrderComparator(config.getSystemWorkspaceName()));
    for (int i = 0; i < wsEntries.size(); i++) {
      initWorkspace(repository, (WorkspaceEntry) wsEntries.get(i), i == 0);
    }        
  }
  
  private void initWorkspace(RepositoryImpl repository, WorkspaceEntry wsConfig, boolean isSystem) throws RepositoryException {
    
    WorkspaceContainer workspaceContainer = (WorkspaceContainer) container.getComponentInstance(wsConfig.getName());
    
    UUIDGenerator uuidGen = (UUIDGenerator) container.getComponentInstanceOfType(UUIDGenerator.class);
        
    // [PN] 27.05.06 Refactoring of system session usage 
//    SessionImpl sysSession = new SessionImpl(
//        wsConfig.getName(), AccessManager.systemCredentials(), workspaceContainer);
//    workspaceContainer.registerComponentInstance(sysSession);
    
    if (isSystem) {
      // [PN] 27.05.06 Refactoring of system session usage
      // system session
      //container.registerComponentInstance(sysSession);
      
      NamespaceRegistryImpl nsRegistry = (NamespaceRegistryImpl) container.getComponentInstanceOfType(NamespaceRegistry.class);
      NodeTypeManagerImpl ntManager = (NodeTypeManagerImpl) container.getComponentInstanceOfType(NodeTypeManager.class);
            
      // Load from persistence
      nsRegistry.loadFromStorage();
      ntManager.loadFromStorage();
    }
    
    // Init Root and jcr:system if workspace is system workspace
    if (wsConfig.getAutoInitializedRootNt() != null && !repository.isWorkspaceInitialized(wsConfig.getName())) {
      repository.initWorkspace(wsConfig.getName(), wsConfig.getAutoInitializedRootNt());
    }
    
    // ROOT node shoould be initialized already
    if (repository.isWorkspaceInitialized(wsConfig.getName())) {

      WorkspaceDataManager dataManager = 
        (WorkspaceDataManager) workspaceContainer.getComponentInstanceOfType(WorkspaceDataManager.class);

      LocationFactory locationFactory = (LocationFactory) container.getComponentInstanceOfType(LocationFactory.class);
      QueryHandlerContext context = new QueryHandlerContext(dataManager,
          Constants.ROOT_UUID, locationFactory);

      QueryHandler handler = (QueryHandler) workspaceContainer.getComponentInstanceOfType(QueryHandler.class);
      if (handler != null) {
        try {
          handler.init(context);
        } catch (IOException e) {
          e.printStackTrace();
          throw new RepositoryException(e);
        }
      }
    }
  }
}
