/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.storage.jdbc;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.RepositoryException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlList;
import org.exoplatform.services.jcr.dataflow.persistent.PersistedNodeData;
import org.exoplatform.services.jcr.dataflow.persistent.PersistedPropertyData;
import org.exoplatform.services.jcr.dataflow.persistent.PersistedValueData;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.UUIDGenerator;
import org.exoplatform.services.jcr.storage.WorkspaceStorageConnection;
import org.exoplatform.services.log.ExoLogger;

/**
 * Created by The eXo Platform SARL        .
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: JDBCStorageConnection.java 5834 2006-05-29 12:05:59Z peterit $
 */

abstract public class JDBCStorageConnection extends DBConstants implements WorkspaceStorageConnection {

  protected PreparedStatement findItemById;
  //protected PreparedStatement findNodeById;
  //protected PreparedStatement findPropertyById;
  protected PreparedStatement findItemByPath;
  protected PreparedStatement findChildItemByPath;
  
  //protected PreparedStatement findValuesByTypeAndValue;
  protected PreparedStatement findPropertiesByValue;
  
  protected PreparedStatement findValuesByPropertyId;
  protected PreparedStatement findNodesByParentId;
  protected PreparedStatement findNodesIdsByParentId;  
  protected PreparedStatement findPropertiesByParentId;
  
  protected PreparedStatement findNodesCountByParentId;
  protected PreparedStatement findPropertiesCountByParentId;

  protected PreparedStatement insertItem;
  protected PreparedStatement insertNode;
  protected PreparedStatement insertProperty;
  protected PreparedStatement insertValue;

  protected PreparedStatement updateItem;
  protected PreparedStatement updateNode;
  protected PreparedStatement updateProperty;
  //protected PreparedStatement updateValue;

  protected PreparedStatement deleteItem;
  protected PreparedStatement deleteNode;
  protected PreparedStatement deleteProperty;
  protected PreparedStatement deleteValue;

  protected static Log log = ExoLogger.getLogger("jcr.JDBCStorageConnection");
  protected String containerName;
  protected int containerNameLength = 0;
  protected final Connection dbConnection;
    
  /**
   * @param dbConnection 
   * @throws SQLException
   */
  protected JDBCStorageConnection(Connection dbConnection, String containerName) throws SQLException {
    this.dbConnection = dbConnection;
    prepareStatements();
    dbConnection.setAutoCommit(false);
    this.containerName = containerName; 
    containerNameLength = containerName.length();
    log.debug("JDBCStorageConnection is opened using "+dbConnection.getClass().getName());
  }

  /**
   * Used in SingleDbJDBCConnection class for UUID related queries 
   * @param uuid
   * @return
   */
  protected String id2cid(String uuid) {
    if (this instanceof SingleDbJDBCConnection && uuid != null && uuid.length() == UUIDGenerator.UUID_LENGTH) {
        return containerName+uuid;
    }
    // uuid already prepared for single db
    return uuid;
  }
  
  /**
   * Used in loadXYZRecord methods for extract real UUID from container value.
   * @param uuid
   * @return
   */
  protected String cid2id(String uuid) {
    if (this instanceof SingleDbJDBCConnection && uuid != null && uuid.length() > UUIDGenerator.UUID_LENGTH) {
      return uuid.substring(containerNameLength);
    }
    return uuid;
  }  
  
  // ---------------- SQLException handler -------------------
  
  protected String handleAddSQLException(SQLException e, ItemData item) throws RepositoryException, InvalidItemStateException {
    String message = "["+containerName+"] ";
    String errMessage = e.getMessage();
    String itemInfo = (item.isNode() ? "NODE " : "PROPERTY ") 
      + item.getQPath().getAsString() + " " + item.getUUID() 
      + (errMessage != null ? ". Cause: " + errMessage : "");
    
    // try detect error by foreign key names
    if (errMessage.indexOf(JCR_FK_NODE_PARENT)>=0) {
      // we see that error contains JCR_FK_XNODE_PARENT constraint name, so it's constraint violation... 
      message += "Parent not found for " + itemInfo;
      throw new InvalidItemStateException(message, e);
    } else if (errMessage.indexOf(JCR_FK_NODE_ITEM)>=0) {
      // Foreign key NODE->ITEM tables vioaltion. A item record not found for node created. 
      message += "Item relation in JCR_ITEM table not found (not created). " + itemInfo;
      throw new RepositoryException(message, e);
    } else if (errMessage.indexOf(JCR_FK_PROPERTY_NODE)>=0) {
      // Foreign key PROPERTY->NODE tables vioaltion 
      message += "Parent not found (not created) in JCR_NODE table. " + itemInfo;
      throw new RepositoryException(message, e);
    } else if (errMessage.indexOf(JCR_FK_PROPERTY_ITEM)>=0) {
      // Foreign key PROPERTY->ITEM tables vioaltion 
      message += "Item relation in JCR_ITEM table not found (not created). " + itemInfo;
      throw new RepositoryException(message, e);
    } else if (errMessage.indexOf(JCR_FK_VALUE_PROPERTY)>=0) {
      // Foreign key VALUE->PROPERTY tables vioaltion 
      message += "Property not found (not created) in JCR_PROPERTY table for values created. " + itemInfo;
      throw new RepositoryException(message, e);
    }
    // try detect integrity violation
    RepositoryException ownException = null;
    try {
      NodeData parent = (NodeData) getItemData(item.getParentUUID());
      if (parent != null) {
        // have a parent
        try { 
          ItemData me = getItemData(item.getUUID());
          if (me != null) {
            // item already exists
            message += "Item already exists in storage: " + itemInfo;
            ownException = new RepositoryException(message, e);
            throw ownException;
          }
        } catch(Exception ep) {
          // item not found or other things but error of item reading
          if (ownException != null) throw ownException;
        }
        message += "Error add " + itemInfo;
        ownException = new RepositoryException(message, e);
        throw ownException;
      }
    } catch(Exception ep) {
      // no parent or error access it
      if (ownException != null) throw ownException;
    }
    message += "Parent not found (may be deleted by other session) for " + itemInfo;
    throw new InvalidItemStateException(message, e);
  }
  
  protected String handleDeleteSQLException(SQLException e, ItemData item) throws RepositoryException, InvalidItemStateException {
    String message = "["+containerName+"] ";
    String errMessage = e.getMessage();
    String itemInfo = (item.isNode() ? "NODE " : "PROPERTY ") 
      + item.getQPath().getAsString() + " " + item.getUUID() 
      + (errMessage != null ? ". Cause: " + errMessage : "");
    
    // try detect error by foreign key names
    if (errMessage.indexOf(JCR_FK_NODE_PARENT)>=0) {
      // we see that error contains JCR_FK_MNODEPARENT constraint name, so it's constraint violation... 
      message += "Can't delete parent till child exists " + itemInfo;
      throw new InvalidItemStateException(message, e);
    } else if (errMessage.indexOf(JCR_FK_NODE_ITEM)>=0) {
      // Foreign key NODE->ITEM tables vioaltion. 
      message += "Can't delete item till related node exists in JCR_NODE table " + itemInfo;
      throw new RepositoryException(message, e);
    } else if (errMessage.indexOf(JCR_FK_PROPERTY_NODE)>=0) {
      // Foreign key PROPERTY->NODE tables vioaltion 
      message += "Can't delete node till child property(ies) exists in JCR_PROPERTY table " + itemInfo;
      throw new RepositoryException(message, e);
    } else if (errMessage.indexOf(JCR_FK_PROPERTY_ITEM)>=0) {
      // Foreign key PROPERTY->ITEM tables vioaltion 
      message += "Can't delete item till related property exists in JCR_PROPERTY table " + itemInfo;
      throw new RepositoryException(message, e);
    } else if (errMessage.indexOf(JCR_FK_VALUE_PROPERTY)>=0) {
      // Foreign key VALUE->PROPERTY tables vioaltion 
      message += "Can't delete property till related property values exists in JCR_VALUE table " + itemInfo;
      throw new RepositoryException(message, e);
    }
    message += "Error of item delete " + itemInfo;
    throw new RepositoryException(message, e);
  }
  
  protected String handleUpdateSQLException(SQLException e, ItemData item) throws RepositoryException, InvalidItemStateException {
    String message = "["+containerName+"] ";
    String errMessage = e.getMessage();
    String itemInfo = (item.isNode() ? "NODE " : "PROPERTY ") 
      + item.getQPath().getAsString() + " " + item.getUUID() 
      + (errMessage != null ? ". Cause: " + errMessage : "");
    
    // try detect error by foreign key names
    if (errMessage.indexOf(JCR_FK_VALUE_PROPERTY)>=0) {
      // Foreign key VALUE->PROPERTY tables vioaltion 
      message += "Property not found (not created) in JCR_PROPERTY table for values updated. " + itemInfo;
      throw new RepositoryException(message, e);
    }
    // try detect integrity violation
    RepositoryException ownException = null;
    try { 
      ItemData me = getItemData(item.getUUID());
      if (me != null) {
        // item already exists
        message += "Item exists but update error " + itemInfo;
        ownException = new RepositoryException(message, e);
        throw ownException;
      }
    } catch(Exception ep) {
      // item not found or other things but error of item reading
      if (ownException != null) throw ownException;
    }
    message += "Item not found (may be deleted by other session) " + itemInfo;
    throw new InvalidItemStateException(message, e);
  }
  
  // ---------------- WorkspaceStorageConnection ------------- 
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#add(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public void add(NodeData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      addNodeRecord(data);
    } catch (SQLException e) {
      handleAddSQLException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#add(org.exoplatform.services.jcr.datamodel.PropertyData)
   */
  public void add(PropertyData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      addPropertyRecord(data);
    } catch (SQLException e) {
      handleAddSQLException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#delete(org.exoplatform.services.jcr.datamodel.ItemData)
   */
  public void delete(ItemData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      if (data.isNode()) {
        if (deleteNode(data.getUUID()) <= 0) {
          // TODO [PN] it's error state, as we actually didn't delete the node (with sub-nodes)
          log.warn("Error state, the node (with sub-nodes) actually not deleted " + data.getQPath().getAsString());
        }      
      } else {
        if (deleteProperty(data.getUUID()) <= 0) {
          // TODO [PN] it's error state, as we actually didn't delete the property
          log.warn("Error state, the property item actually not deleted " + data.getQPath().getAsString());
        }         
      }
    } catch (SQLException e) {
      handleDeleteSQLException(e, data);
    }
  }
  
  protected int deleteNode(String uuid) throws SQLException {
    int deleted = 0;

    // deleting child nodes 
    ResultSet childNodes = findChildNodesIdsByParentUUID(uuid);
    List<String> childNodesIds = new ArrayList<String>(); 
    while (childNodes.next()) {
      childNodesIds.add(childNodes.getString(COLUMN_ID));
    }
    childNodes.close();
    
    for (String childNodesId: childNodesIds) {
      int nc = deleteNode(childNodesId);
      if (nc <= 0) {
        // TODO [PN] it's error state, as we actually didn't delete the child node
        log.warn("Error state, the node item actually not deleted " + childNodes.getString(COLUMN_PATH));
      }
      deleted += nc;
    }
    
    // deleting child properties
    ResultSet childProperties = findChildPropertiesByParentUUID(uuid);
    while (childProperties.next()) {
      if (deleteProperty(childProperties.getString(COLUMN_PID)) <= 0) {
        // TODO [PN] it's error state, as we actually didn't delete the property
        log.warn("Error state, the property item actually not deleted " + childProperties.getString(COLUMN_PATH));
      }
    }
    childProperties.close();
    
    // deleting from NODE table 
    if (deleteNodeByUUID(uuid) <= 0) {
      // TODO [PN] it's error state, as we actually didn't delete the node
      log.warn("Error state, the node actually not deleted " + uuid);
    }
    
    // deleting from ITEM table 
    int nc = deleteItemByUUID(uuid);
    if (nc <= 0) {
      // TODO [PN] it's error state, as we actually didn't delete a item corresponding the node
      log.warn("Error state, a item corresponding the node actually not deleted " + uuid);
    }
    deleted += nc;
    return deleted;
  }
  
  protected int deleteProperty(String uuid) throws SQLException {
    // deleting values
    deleteValueByUUID(uuid);
    // deleting properties
    if (deletePropertyByUUID(uuid) <= 0) {
      log.warn("Error state, the property actually not deleted " + uuid);
    }
    // deleting item
    return deleteItemByUUID(uuid);
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#update(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public void update(NodeData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      // version update
      updateItemVersionByUUID(data.getPersistedVersion(), data.getUUID());
      // order numb update
      updateNodeOrderNumbByUUID(data.getOrderNumber(), data.getUUID());      
    } catch (SQLException e) {
      handleUpdateSQLException(e, data);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#update(org.exoplatform.services.jcr.datamodel.PropertyData)
   */
  public void update(PropertyData data) throws RepositoryException, UnsupportedOperationException, InvalidItemStateException, IllegalStateException {
    checkIfOpened();
    try {
      // version update
      updateItemVersionByUUID(data.getPersistedVersion(), data.getUUID());

      // update type
      updatePropertyTypeByUUID(data.getType(), data.getUUID());
      
      // deleting property values
      deleteValueByUUID(data.getUUID());
      // and creating new
      addValueRecords(data);
    } catch (SQLException e) {
      handleUpdateSQLException(e, data);
    }
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildNodesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public int getChildNodesCount(NodeData nodeData) throws RepositoryException {
    checkIfOpened();
    try {
      return getChildNodesCountByParentUUID(nodeData.getParentUUID()).getInt(1);
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildNodesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<NodeData> getChildNodesData(NodeData parent) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    try {
      ResultSet node = findChildNodesByParentUUID(parent.getUUID());
      List<NodeData> childrens = new ArrayList<NodeData>(); 
      while(node.next()) {
        String id = node.getString(COLUMN_NID);
        if (id != null) {
          //ResultSet item = findItemByUUID(id);
          childrens.add(loadNodeRecord(node));
        } else {
          throw new RepositoryException("FATAL: Not found child node for parent "+parent.getQPath().getAsString()
              + ", but child item found " + node.getString(COLUMN_PATH) + " " + node.getString(COLUMN_ID));
        }
      }
      return childrens;      
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildPropertiesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public int getChildPropertiesCount(NodeData parent) throws RepositoryException {
    checkIfOpened();
    try {
      return getChildPropertiesCountByParentUUID(parent.getUUID()).getInt(1);
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildPropertiesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<PropertyData> getChildPropertiesData(NodeData parent) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    try {
      ResultSet prop = findChildPropertiesByParentUUID(parent.getUUID());
      List<PropertyData> children = new ArrayList<PropertyData>(); 
      while(prop.next()) {
        String id = prop.getString(COLUMN_PID);
        //ResultSet item = findItemByUUID(id);
        if (id != null) {
          //ResultSet item = findItemByUUID(id);
          children.add(loadPropertyRecord(prop));
        } else {
          throw new RepositoryException("FATAL: Not found child property for parent "+parent.getQPath().getAsString()
              + ", but child item found " + prop.getString(COLUMN_PATH) + " " + prop.getString(COLUMN_ID));
        }
      }
      return children;
      
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getItemData(org.exoplatform.services.jcr.datamodel.InternalQPath)
   */
  public ItemData getItemData(InternalQPath qPath) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    ResultSet item = null;
    try {
      item = findItemByPath(qPath.getAsString());
      if(item.next()) {
        return itemData(item);
      }
      return null;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    } finally {
      try {
        if (item != null) 
          item.close();
      } catch(SQLException e) {
        log.error("getItemData() Error close resultset " + e.getMessage());
      }
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getItemData(java.lang.String)
   */
  public ItemData getItemData(String uuid) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    ResultSet item = null;
    try {
      item = findItemByUUID(uuid);
      if(item.next()) {
        return itemData(item);
      }
      return null;
    } catch (SQLException e) {
      throw new RepositoryException("getItemData() error", e);
    } finally {
        try {
          if (item != null) 
            item.close();
        } catch(SQLException e) {
          log.error("getItemData() Error close resultset " + e.getMessage());
        }
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getReferencesData(java.lang.String)
   */
  public List<PropertyData> getReferencesData(String nodeUUID) throws RepositoryException, IllegalStateException {
    checkIfOpened();
    try {
      ResultSet refProps = findReferences(nodeUUID);
      List<PropertyData> references = new ArrayList<PropertyData>(); 
      while(refProps.next()) {
        String id = refProps.getString(COLUMN_PID);
        if (id != null) {
          // 6.7.21.2 mix:referenceable
          // we know that REFERENCE property is single-valued (Multiple false)
          byte[] data = refProps.getBytes(COLUMN_VDATA);
          int orderNum = refProps.getInt(COLUMN_VORDERNUM);
          List<ValueData> value = new ArrayList<ValueData>();
          PersistedValueData record = new PersistedValueData(data, orderNum); 
          value.add(record);
          
//          log.info("getReferencesData() >>> ");
//          for (ValueData v: value) {
//            if (v != null)
//              log.info("getReferencesData(): v:" + new String(v.getAsByteArray()));
//            else 
//              log.info("getReferencesData(): null v ");
//          }
//          log.info("getReferencesData() <<< ");
          
          PropertyData refProp = new PersistedPropertyData(
              cid2id(id), 
              InternalQPath.parse(refProps.getString(COLUMN_PATH)), 
              cid2id(refProps.getString(COLUMN_PPARENTID)), 
              refProps.getInt(COLUMN_VERSION),
              refProps.getInt(COLUMN_PTYPE),
              refProps.getBoolean(COLUMN_PMULTIVALUED),
              value, 
              null // TODO temporary
              );
          references.add(refProp);
        } else {
          throw new RepositoryException("FATAL: Not found references property for UUID "+nodeUUID
              + ", but references item found " + refProps.getString(COLUMN_PATH) + " " + cid2id(refProps.getString(COLUMN_ID)));
        }
      }
      return references;
    } catch (SQLException e) {
      e.printStackTrace();
      throw new RepositoryException(e);
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#isOpened()
   */
  public boolean isOpened() {
    try {
      return !dbConnection.isClosed();
    } catch (SQLException e) {
      log.error(e);
      return false;
    }
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#rollback()
   */
  public final void rollback() throws IllegalStateException, RepositoryException {
    checkIfOpened();
    try {
      dbConnection.rollback();
      dbConnection.close();
    } catch (SQLException e) {
      throw new RepositoryException(e);
    }
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#commit()
   */
  public final void commit() throws IllegalStateException, RepositoryException {
    checkIfOpened();
    try {
      dbConnection.commit();
      dbConnection.close();
    } catch (SQLException e) {
      throw new RepositoryException(e);
    }
  }

  // ------------------ Private methods ---------------
  
  /**
   * @param itemRecord
   * @return
   * @throws RepositoryException
   * @throws SQLException
   */
  private ItemData itemData(ResultSet itemRecord) throws RepositoryException, SQLException {
    String id = itemRecord.getString(COLUMN_ID);
    
    boolean isNode = itemRecord.getString(COLUMN_NID) != null;
    boolean isProperty = itemRecord.getString(COLUMN_PID) != null;
    
    if (isNode && isProperty) {
      // inconsistent data stored in db
      throw new RepositoryException("FATAL: Item stored as node and as property "+itemRecord.getString(COLUMN_PATH) + " " + id);
    }
    
    if (!isNode && !isProperty) {
      // no data stored for item - node or property
      throw new RepositoryException("FATAL: Not found corresponding node nor property for item "+itemRecord.getString(COLUMN_PATH));
    }
   
    if (isNode) {
      return loadNodeRecord(itemRecord);
    } else {
      // property
      return loadPropertyRecord(itemRecord);
      // found item record but no prop
      //throw new RepositoryException("FATAL: Not found corresponding property for item "+itemRecord.getString(COLUMN_PATH));      
    }   
  }
  
  /**
   * @throws IllegalStateException if connection is closed
   */
  private void checkIfOpened() throws IllegalStateException {
    if(!isOpened())
      throw new IllegalStateException("Connection is closed");
  }
  
  /**
   * Prepared statemens at statr time as it recommends by JDBC API doc
   * @throws SQLException
   */
  protected final void prepareStatements() throws SQLException {
    //System.out.println("Connection: "+dbConnection);  
    findItemById = dbConnection.prepareStatement(FIND_ITEM_BY_ID);
    //findNodeById = dbConnection.prepareStatement(FIND_NODE_BY_ID);
    //findPropertyById = dbConnection.prepareStatement(FIND_PROPERTY_BY_ID);
    findItemByPath = dbConnection.prepareStatement(FIND_ITEM_BY_PATH);
    findChildItemByPath = dbConnection.prepareStatement(FIND_CHILDITEM_BY_PATH);
    
    //findNodeByPath = dbConnection.prepareStatement(FIND_NODE_BY_PATH);
    //findPropertyByPath = dbConnection.prepareStatement(FIND_PROPERTY_BY_PATH);
    
    //findValuesByTypeAndValue = dbConnection.prepareStatement(FIND_PROPSVALUES_BY_TYPE_AND_VALUE);
    findPropertiesByValue = dbConnection.prepareStatement(FIND_PROPERTIES_BY_VALUE);
    
    findValuesByPropertyId = dbConnection.prepareStatement(FIND_VALUES_BY_PROPERTYID);
    
    findNodesByParentId = dbConnection.prepareStatement(FIND_NODES_BY_PARENTID);
    findNodesIdsByParentId = dbConnection.prepareStatement(FIND_NODES_IDS_BY_PARENTID);
    findPropertiesByParentId = dbConnection.prepareStatement(FIND_PROPERTIES_BY_PARENTID);
    
    findNodesCountByParentId = dbConnection.prepareStatement(FIND_NODESCOUNT_BY_PARENTID);
    findPropertiesCountByParentId = dbConnection.prepareStatement(FIND_PROPERTIESCOUNT_BY_PARENTID);

    insertItem = dbConnection.prepareStatement(INSERT_ITEM);
    insertNode = dbConnection.prepareStatement(INSERT_NODE);
    insertProperty = dbConnection.prepareStatement(INSERT_PROPERTY);
    insertValue = dbConnection.prepareStatement(INSERT_VALUE);

    updateItem = dbConnection.prepareStatement(UPDATE_ITEM);
    updateNode = dbConnection.prepareStatement(UPDATE_NODE);
    updateProperty = dbConnection.prepareStatement(UPDATE_PROPERTY);
    //updateValue = dbConnection.prepareStatement(UPDATE_VALUE);

    deleteItem = dbConnection.prepareStatement(DELETE_ITEM);
    deleteNode = dbConnection.prepareStatement(DELETE_NODE);
    deleteProperty = dbConnection.prepareStatement(DELETE_PROPERTY);
    deleteValue = dbConnection.prepareStatement(DELETE_VALUE);
  }
  
  protected PersistedNodeData loadNodeRecord(ResultSet item) throws RepositoryException, SQLException {
    
    String cNID = item.getString(COLUMN_NID);
    String cNPARENTID = item.getString(COLUMN_NPARENTID);
    int cVERSION = item.getInt(COLUMN_VERSION);
    int cNORDERNUM = item.getInt(COLUMN_NORDERNUM);
    InternalQPath qpath = InternalQPath.parse(item.getString(COLUMN_PATH));
    
    // PRIMARY
    InternalQPath ptPath = InternalQPath.makeChildPath(qpath, Constants.PRIMARY_TYPE_URI);
    
    ResultSet ptProp = findItemByPath(cNID, ptPath.getAsString());

    if (!ptProp.next())
      throw new RepositoryException("FATAL ERROR primary type record not found "
          + ptPath.getAsString()+", container "+this.containerName);
    
    ResultSet ptValue = findValuesByPropertyId(ptProp.getString(COLUMN_PID));

    if (!ptValue.next())
      throw new RepositoryException("FATAL ERROR primary type value not found "
          + ptPath.getAsString()+", container "+this.containerName);

    InternalQName ptName = InternalQName.parse(new String(ptValue.getBytes(COLUMN_DATA)));

    // MIXIN
    InternalQPath mtPath = InternalQPath.makeChildPath(qpath, Constants.MIXIN_TYPE_URI);
    // [PN] 22.05.06
    //ResultSet mtProp = findItemByPath(mtPath.getAsString());
    ResultSet mtProp = findItemByPath(cNID, mtPath.getAsString());
    
    InternalQName[] mixinNames = new InternalQName[0];
    if(mtProp.next()) {
      List<byte[]> mts = new ArrayList<byte[]>();
      ResultSet mtValues = findValuesByPropertyId(mtProp.getString(COLUMN_PID));
      while (mtValues.next()) {
        mts.add(mtValues.getBytes(COLUMN_DATA));
      }
      mixinNames = new InternalQName[mts.size()];
      for (int i = 0; i < mts.size(); i++) {
        mixinNames[i] = InternalQName.parse(new String(mts.get(i)));
      }
    }

    // ACL
    AccessControlList acl;
    if (isAccessControllable(mixinNames)) {
      
      InternalQPath ownerPath = InternalQPath.makeChildPath(qpath,
          Constants.NODE_OWNER_URI);
      
      PropertyData ownerData = (PropertyData) getItemData(ownerPath);
      
      InternalQPath permPath = InternalQPath.makeChildPath(qpath,
          Constants.NODE_PERM_URI);
   
      PropertyData permData = (PropertyData) getItemData(permPath);
      
      acl = new AccessControlList(ownerData, permData);
    } else {
      acl = new AccessControlList(true);
    }
          
    return new PersistedNodeData(cid2id(cNID), 
        qpath, 
        cid2id(cNPARENTID),
        cVERSION,
        cNORDERNUM,
        ptName, mixinNames, acl);
        
  }
  
  private boolean isAccessControllable(InternalQName[] mixinNames) {
    for (int i = 0; i < mixinNames.length; i++) {
      if (mixinNames[i].getAsString().equals(Constants.ACCESS_TYPE_URI)) {
        return true;
      }
    }
    return false;
  }  
  
    
  protected PersistedPropertyData loadPropertyRecord(ResultSet item) throws RepositoryException, SQLException {
    
    ResultSet valueRecords = findValuesByPropertyId(item.getString(COLUMN_PID));
    
    List<ValueData> values = new ArrayList<ValueData>();
    while(valueRecords.next()) {
      byte[] data = valueRecords.getBytes(COLUMN_DATA);
      int orderNum = valueRecords.getInt(COLUMN_ORDERNUM);
      PersistedValueData record = new PersistedValueData(data, orderNum); 
      values.add(record);
    }
    
//    log.info("loadPropertyRecord() >>> ");
//    if (item.getString(COLUMN_PATH).equals("[]:1[]testdata:1[]docViewTest:1[]bigNode:1[]propName0:1")) {
//      log.info(" >>> ");
//    }
//    for (ValueData v: values) {
//      if (v != null)
//        log.info("loadPropertyRecord(): v:" + new String(v.getAsByteArray()) + "  " + item.getString(COLUMN_PATH));
//      else 
//        log.info("loadPropertyRecord(): null v  " + item.getString(COLUMN_PATH));
//    }
//    log.info("loadPropertyRecord() <<< ");

    return new PersistedPropertyData(
        cid2id(item.getString(COLUMN_PID)), 
        InternalQPath.parse(item.getString(COLUMN_PATH)),
        cid2id(item.getString(COLUMN_PPARENTID)),
        item.getInt(COLUMN_VERSION),
        item.getInt(COLUMN_PTYPE),
        item.getBoolean(COLUMN_PMULTIVALUED),
        values, 
        null // TODO temporary
        );
  }
  
  // ---- Data access methods (query wrappers) to override in concrete connection ------
  
  protected abstract void addNodeRecord(NodeData data) throws SQLException;

  protected abstract void addPropertyRecord(PropertyData prop) throws SQLException;
  
  protected abstract void addValueRecords(PropertyData data) throws SQLException;

  protected abstract ResultSet findItemByPath(String path) throws SQLException;
  
  // [PN] find child item by path for parent identified by parentId 
  protected abstract ResultSet findItemByPath(String parentId, String path) throws SQLException;

  protected abstract ResultSet findItemByUUID(String uuid) throws SQLException;

  //protected abstract ResultSet findNodeByPath(String path) throws SQLException;

  //protected abstract ResultSet findNodeByUUID(String uuid) throws SQLException;

  //protected abstract ResultSet findPropertyByPath(String path) throws SQLException;

  //protected abstract ResultSet findPropertyByUUID(String uuid) throws SQLException;
  
  protected abstract ResultSet findValuesByPropertyId(String propertyId) throws SQLException;
  
  protected abstract ResultSet findReferences(String uuid) throws SQLException;
  
  protected abstract ResultSet getChildNodesCountByParentUUID(String parentUUID) throws SQLException;
  
  protected abstract ResultSet getChildPropertiesCountByParentUUID(String parentUUID) throws SQLException;
  
  protected abstract ResultSet findChildNodesByParentUUID(String parentUUID) throws SQLException;
  
  // [PN] 22.05.06 For deleteNode() method
  protected abstract ResultSet findChildNodesIdsByParentUUID(String parentUUID) throws SQLException;
  
  protected abstract ResultSet findChildPropertiesByParentUUID(String parentUUID) throws SQLException;
  
  protected abstract int deleteItemByUUID(String uuid) throws SQLException;
  protected abstract int deleteNodeByUUID(String uuid) throws SQLException;
  protected abstract int deletePropertyByUUID(String uuid) throws SQLException;
  protected abstract int deleteValueByUUID(String uuid) throws SQLException;
  
  protected abstract int updateItemVersionByUUID(int versionValue, String uuid) throws SQLException;
  protected abstract int updateNodeOrderNumbByUUID(int orderNumb, String uuid) throws SQLException;
  protected abstract int updatePropertyTypeByUUID(int type, String uuid) throws SQLException;
}
