/***************************************************************************
 * Copyright 2001-2006 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.storage.rdb.container;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.ItemExistsException;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AccessControlList;
import org.exoplatform.services.jcr.datamodel.IllegalNameException;
import org.exoplatform.services.jcr.datamodel.IllegalPathException;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.dataflow.data.ItemDataFactory;
import org.exoplatform.services.jcr.impl.storage.rdb.container.data.ItemRecord;
import org.exoplatform.services.jcr.impl.storage.rdb.container.data.NodeRecord;
import org.exoplatform.services.jcr.impl.storage.rdb.container.data.PropertyRecord;
import org.exoplatform.services.jcr.impl.storage.rdb.container.data.ValueRecord;
import org.exoplatform.services.jcr.storage.WorkspaceStorageConnection;
import org.exoplatform.services.log.ExoLogger;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;

/**
 * Created by The eXo Platform SARL 
 * Author : Peter Nedonosko
 *          peter.nedonosko@exoplatform.com.ua 
 * 20.03.2006
 * 
 * @version $Id: HibernateStorageConnection.java 5307 2006-04-27 13:55:57Z peterit $
 */
abstract public class HibernateStorageConnection implements WorkspaceStorageConnection {

  static public SimpleDateFormat LOG_TIME_FORMAT = new SimpleDateFormat("HH:mm:ss SSSS");

  protected static Log log = ExoLogger.getLogger("jcr.HibernateStorageConnection");
  protected SessionFactory sessionFactory = null;
  
  protected Session hibernateSession = null;
  protected Transaction tx = null;
  
  protected long timeCreated = 0;
  
  protected String sourceName = "?";
  
  /** Constructor for inheritance only */
  protected HibernateStorageConnection(String sourceName,
      SessionFactory sessionFactory) {
    
    this.sessionFactory = sessionFactory;
    this.timeCreated = System.currentTimeMillis();
    this.sourceName = sourceName;
    
    open();
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getTimeCreated()
   */
  public long getTimeCreated() {
    return timeCreated;
  }

  protected void toLog(String message) {
    toLog(message, 0);
  }

  protected void toLog(String message, long startTime) {
    String ts = "HibernateStorageConnection [" + LOG_TIME_FORMAT.format(new Date()) + "]";
    if (startTime > 0) {
      ts += ":[" + (System.currentTimeMillis() - startTime) + "] ";
    } else {
      ts += " ";
    }
    if (message != null && message.length() > 80)
      message = message.substring(0, 80);
    // log.info(ts + message);
  }

  private void checkTransaction() throws RepositoryException {
    if (tx == null || !tx.isActive())
      throw new RepositoryException("Transaction is not opened. Call beginBatch");
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#close()
   */
  public void close() {
    hibernateSession.close();
    hibernateSession = null;
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getItemData(org.exoplatform.services.jcr.datamodel.InternalQPath)
   */
  public ItemData getItemData(InternalQPath qPath) throws RepositoryException {
    try {
      checkTransaction();

      ItemRecord data = findItemByPath(qPath.getAsString(), hibernateSession);
      if (data != null && (data instanceof NodeRecord)) {
        initNode((NodeRecord) data);
      }
      
      return ItemDataFactory.makeItemData(data);
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".getItemData(InternalQPath qPath) failed: <" + qPath.getAsString() + "> "
        + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getItemData(java.lang.String)
   */
  public ItemData getItemData(String uuid) throws RepositoryException {
    try {
      checkTransaction();

      ItemRecord data = findItemByUUID(uuid, hibernateSession);
      if (data != null && (data instanceof NodeRecord)) {
        initNode((NodeRecord) data);
      }

      return ItemDataFactory.makeItemData(data);
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".getItemData(String uuid) failed: <" + uuid + "> "
        + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException (errMessage, e);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildPropertiesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public int getChildPropertiesCount(NodeData parentData) throws RepositoryException {
    
    try {
      checkTransaction();
      
      return getChildPropertiesCountByParentUUID(parentData.getUUID(), hibernateSession);
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".getChildPropertiesCount(NodeData parentData) failed: <" + parentData.getQPath().getAsString() + ", " + 
        parentData.getUUID() + "> " + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildNodesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */  
  public int getChildNodesCount(NodeData parentData) throws RepositoryException {
    
    try {
      checkTransaction();
    
      return getChildNodesCountByParentUUID(parentData.getUUID(), hibernateSession);
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".getChildNodesCount(NodeData parentData) failed: <" + parentData.getQPath().getAsString() + ", " + 
        parentData.getUUID() + "> " + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }
  }
  
  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildNodesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<NodeData> getChildNodesData(NodeData parentData) throws RepositoryException {
   
    try {
      checkTransaction();

      List<NodeData> childNodes = new ArrayList<NodeData>();
      List cn = getChildNodesByParentUUID(parentData.getUUID(), hibernateSession);
      if (cn != null && cn.size() > 0) {
        for (Object cno : cn) {
          if (cno instanceof NodeRecord) {
            initNode((NodeRecord) cno);
            childNodes.add(ItemDataFactory.makeNodeData((NodeRecord) cno));
          }
        }
      }
      return childNodes;
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".getChildNodesData(NodeData parentData) failed: <"
          + parentData.getQPath().getAsString() + ", " + parentData.getUUID() + "> " + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getChildPropertiesData(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public List<PropertyData> getChildPropertiesData(NodeData parentData) throws RepositoryException {
    
    try {
      checkTransaction();

      List<PropertyData> childProperties = new ArrayList<PropertyData>();
      List cp = getChildPropertiesByParentUUID(parentData.getUUID(), hibernateSession);
      if (cp != null && cp.size() > 0) {
        for (Object cpo : cp) {
          if (cpo instanceof PropertyRecord) {
            PropertyData rdbData = ItemDataFactory.makePropertyData((PropertyRecord) cpo);
            childProperties.add(rdbData);
          }
        }
        // [PN] 05.04.06
        //return childProperties;
      }
      return childProperties; // properties must be one or more or null will returned
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".getChildPropertiesData(NodeData parentData) failed: <"
          + parentData.getQPath().getAsString() + ", " + parentData.getUUID() + "> "
          + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }
  }
  
  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#add(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public void add(NodeData data) throws RepositoryException, ItemExistsException,
      InvalidItemStateException {

    try {
      checkTransaction();

      addNodeRecord(data, hibernateSession);
      
      // [PN] 03.04.06 return persisted version
      //return nodeRecord.getPersistedVersion();
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".add(NodeData data) failed: <"
          + data.getQPath().getAsString() + ", " + data.getUUID() + "> "
          + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }
  }


  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#add(org.exoplatform.services.jcr.datamodel.PropertyData)
   */
  public void add(PropertyData prop) throws RepositoryException, ItemExistsException,
      InvalidItemStateException {

    try {
      checkTransaction();

      NodeRecord parent = findParentNodeRecord(prop);
      PropertyRecord propertyRecord = addPropertyRecord(prop, parent, hibernateSession);
      
      // [PN] 03.04.06 return persisted version
      //return propertyRecord.getPersistedVersion();
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".add(PropertyData prop) failed: <" + prop.getQPath().getAsString()
          + ", " + prop.getUUID() + "> " + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }    
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#update(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  public void update(NodeData node) throws RepositoryException,
      ItemExistsException {

    try {
      checkTransaction();
  
      NodeRecord nodeRecord = findNodeByUUID(node.getUUID(), hibernateSession);
      nodeRecord.setOrderNumber(node.getOrderNumber());
      nodeRecord.setPath(node.getQPath().getAsString());
      hibernateSession.persist(nodeRecord);
      
      // [PN] 03.04.06 return persisted version
      //return nodeRecord.getPersistedVersion();
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".update(NodeData node) failed: <" + node.getQPath().getAsString()
          + ", " + node.getUUID() + "> " + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }        
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#update(org.exoplatform.services.jcr.datamodel.PropertyData)
   */
  public void update(PropertyData prop) throws RepositoryException {

    try {
      checkTransaction();

      PropertyRecord propertyRecord = findPropertyByUUID(prop.getUUID(), hibernateSession);

      List oldValues = propertyRecord.getValues();
      for (int i = 0; i < oldValues.size(); i++)
        hibernateSession.delete(oldValues.get(i));

      propertyRecord.setType(prop.getType());
      propertyRecord.setValues(prop.getValues());
      propertyRecord.setMultiValued(prop.isMultiValued());

      hibernateSession.persist(propertyRecord);
      
      // [PN] 03.04.06 return persisted version
      //return propertyRecord.getPersistedVersion();
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".update(PropertyData prop) failed: <"
          + prop.getQPath().getAsString() + ", " + prop.getUUID() + "> " + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#delete(org.exoplatform.services.jcr.datamodel.ItemData)
   */
  public void delete(ItemData data) throws RepositoryException, UnsupportedOperationException,
      InvalidItemStateException {

    try {
      checkTransaction();

      ItemData itemRecord = findItemByUUID(data.getUUID(), hibernateSession);
      if (itemRecord != null) {
        hibernateSession.delete(itemRecord);
      }
      // log.debug("RDB Container removed " + data.getQPath().getAsString());
      // [PN] 03.04.06 return persisted version
      //return itemRecord.getPersistedVersion();
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".update(PropertyData prop) failed: <"
          + data.getQPath().getAsString() + ", " + data.getUUID() + "> " + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }    
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#getReferencesData(java.lang.String)
   */
  public List<PropertyData> getReferencesData(String nodeUUID) throws RepositoryException {

    try {
      checkTransaction();
      
      List<PropertyData> references = new ArrayList<PropertyData>();
      List vrefs = findReferences(nodeUUID, hibernateSession);
      if (vrefs != null && vrefs.size() > 0) {
        for (Object vref : vrefs) {
          if (vref instanceof ValueRecord) {
            ValueRecord rref = (ValueRecord) vref;
            if (rref.getProperty().getType() == PropertyType.REFERENCE) {
              PropertyData rdbData = ItemDataFactory.makePropertyData(rref.getProperty());
              references.add(rdbData);
            }
          }
        }
      }            
      return references;
    } catch (RuntimeException e) {
      String errMessage = getClass().getName() + ".getReferencesData(String nodeUUID) failed: <"
          + nodeUUID + "> " + e.getMessage();
      log.error(errMessage, e);
      throw new RepositoryException(errMessage, e);
    }        
  }

  protected NodeRecord findParentNodeRecord(ItemData data) throws RepositoryException,
      InvalidItemStateException {
    NodeRecord parentNodeRecord = null;
    if (!data.getQPath().getAsString().equals(Constants.ROOT_URI)) {
      String parentPath = data.getQPath().makeParentPath().getAsString();
      log.debug("Finding parent record " + parentPath);

      parentNodeRecord = findNodeByPath(parentPath, hibernateSession);

      if (parentNodeRecord == null)
        throw new InvalidItemStateException("Parent not found for " + data.getQPath().getAsString()
            + " probably was removed by other client. "
            + parentPath);
    }
    return parentNodeRecord;
  }

  /**
   * Inits node's types(primary & mixin)
   * 
   * @param nodeRecord
   * @param hibernateSession
   * @throws HibernateException
   * @throws RepositoryException
   */
  protected void initNode(NodeRecord data) throws HibernateException,
      RepositoryException, IllegalPathException, IllegalNameException {

    // PRIMARY
    InternalQPath propPath = InternalQPath.makeChildPath(data.getQPath(),
        Constants.PRIMARY_TYPE_URI);
    PropertyRecord ptProp = findPropertyByPath(propPath.getAsString(),
        hibernateSession);

    if (ptProp == null)
      throw new RepositoryException("FATAL ERROR Not found primary type "
          + propPath.getAsString());

    ValueRecord ptValue = (ValueRecord) ptProp.getValues().get(0);
    InternalQName ptName = InternalQName.parse(new String(ptValue
        .getAsByteArray()));
    data.setPrimaryType(ptName);

    // MIXIN
    InternalQPath mtPath = InternalQPath.makeChildPath(data.getQPath(),
        Constants.MIXIN_TYPE_URI);
    PropertyRecord mtProp = findPropertyByPath(mtPath.getAsString(),
        hibernateSession);

    InternalQName[] mixinNames;
    if (mtProp == null) {
      mixinNames = new InternalQName[0];
    } else {
      List ptValues = mtProp.getValues();
      mixinNames = new InternalQName[ptValues.size()];
      for (int i = 0; i < ptValues.size(); i++) {
        ValueRecord val = (ValueRecord) ptValues.get(i);
        mixinNames[i] = InternalQName.parse(new String(val.getAsByteArray()));
      }
    }
    data.setMixinTypes(mixinNames);

    // ACL
    if (isAccessControllable(mixinNames)) {
      InternalQPath ownerPath = InternalQPath.makeChildPath(data.getQPath(),
          Constants.NODE_OWNER_URI);
      PropertyRecord ownerRecord = findPropertyByPath(ownerPath.getAsString(),
          hibernateSession);
      InternalQPath permPath = InternalQPath.makeChildPath(data.getQPath(),
          Constants.NODE_PERM_URI);
      PropertyRecord permissionsRecord = findPropertyByPath(permPath
          .getAsString(), hibernateSession);

      data.setChildrenACL(new AccessControlList(ownerRecord, permissionsRecord));
    }
    
  }

  private boolean isAccessControllable(InternalQName[] mixinNames) {
    for (int i = 0; i < mixinNames.length; i++) {
      if (mixinNames[i].getAsString().equals(Constants.ACCESS_TYPE_URI)) {
        return true;
      }
    }
    return false;
  }  
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.storage.WorkspaceStorageConnection#isOpen()
   */
  public boolean isOpened() {
    return !sessionFactory.isClosed();
  }

  public void open() {
    //toLog("beginTransaction() start: " + tx);
    this.hibernateSession = sessionFactory.openSession();
    tx = hibernateSession.beginTransaction();
    //toLog("beginTransaction() fin: " + tx);
  }

  public void commit() {
    toLog("commit(): start: " + tx);
    //long ts = System.currentTimeMillis();
    if (tx != null) {
      if (tx.isActive()) {
        tx.commit();
      }
      tx = null;
    }
    //toLog("commit(): fin: " + tx, ts);
    close();
  }

  public void rollback() {
    //toLog("rollback(): start: " + tx);
    //long ts = System.currentTimeMillis();
    if (tx != null) {
      if (tx.isActive()) {
        tx.rollback();
      }
      tx = null;
    }
    //toLog("rollback(): fin: " + tx, ts);
    close();
  }
  
  protected abstract NodeRecord addNodeRecord(NodeData data,
      Session hibernateSession) throws RepositoryException,
      ItemExistsException, InvalidItemStateException;

  protected abstract PropertyRecord addPropertyRecord(PropertyData prop,
      NodeRecord parent, Session hibernateSession) throws RepositoryException,
      ItemExistsException;

  protected abstract ItemRecord findItemByPath(String path,
      Session hibernateSession);

  protected abstract ItemRecord findItemByUUID(String uuid,
      Session hibernateSession);

  protected abstract NodeRecord findNodeByPath(String path,
      Session hibernateSession);

  protected abstract NodeRecord findNodeByUUID(String uuid,
      Session hibernateSession);

  protected abstract PropertyRecord findPropertyByPath(String path,
      Session hibernateSession);

  protected abstract PropertyRecord findPropertyByUUID(String uuid,
      Session hibernateSession);

  protected abstract List findReferences(String uuid, Session hibernateSession);
  
  protected abstract int getChildNodesCountByParentUUID(String parentUUID, Session hibernateSession);
  
  protected abstract int getChildPropertiesCountByParentUUID(String parentUUID, Session hibernateSession);
  
  protected abstract List getChildNodesByParentUUID(String parentUUID, Session hibernateSession);
  
  protected abstract List getChildPropertiesByParentUUID(String parentUUID, Session hibernateSession);
  
  // experimental
  protected abstract void initIndexes();
  
}
